;;;; geopoint-utils.scm  -*- Scheme -*-
;;;; Kon Lovett, Jul '18
;;;; Kon Lovett, Aug '17
;;;; Kon Lovett, May '17

(module geopoint-utils

(;export
  geopoint-in-closed-polygon?
  intersects?-pnp intersects?-pip)

(import scheme
  (chicken type)
  geopoint
  geopolygon)

(define-type geopoint (struct geopoint))
(define-type geopolygon (or (list-of geopoint) (vector-of geopoint)))

(: geopoint-in-closed-polygon? (geopoint geopolygon #!optional (geopoint geopoint number number -> boolean) --> boolean))
;(: geopoint-in-closed-polygon? (geopoint geopolygon #!optional procedure --> boolean))
(: intersects?-pnp (geopoint geopoint number number --> boolean))
(: intersects?-pip (geopoint geopoint number number --> boolean))

;;

;https://github.com/substack/point-in-polygon
;
(define (geopoint-in-closed-polygon? gp gpoly #!optional (intersects? intersects?-pnp))
  ;test for intersection of ray with every segment of the polygon
  ;start with the "closing" segment, then every [i-i],[i] segment
  (let ((gpoly (make-geopolygon gpoly)))
    (check-geopoint 'geopoint-in-closed-polygon? gp)
    (let ((len (vector-length (check-geopolygon 'geopoint-in-closed-polygon? gpoly)))
          (lat (geopoint-latitude gp))
          (lon (geopoint-longitude gp)) )
      ;assumes an open-poly is "closed" so a closed-poly must be treated as "open"
      (let ((len (if (geopolygon-closed? gpoly) (- len 1) len)))
        (let loop ((i (the fixnum 0)) (j (- len 1)) (poly? #f))
          (if (= i len)
              poly?
              (let ((new-poly
                      (if (intersects? (vector-ref gpoly i) (vector-ref gpoly j) lat lon)
                          (not poly?)
                          poly?)) )
              (loop (+ i 1) i new-poly) ) ) ) ) ) ) )

;;

;; Ray-casting algorithm (http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html)
;; (https://github.com/substack/point-in-polygon)
;
(define (intersects?-pnp pi pj lat lon)
  (let ((lati (geopoint-latitude pi))
        (loni (geopoint-longitude pi))
        (latj (geopoint-latitude pj))
        (lonj (geopoint-longitude pj)) )
    (and (not (eq? (> loni lon) (> lonj lon)))
         (< lat (+ (/ (* (- latj lati) (- lon loni)) (- lonj loni)) lati))) ) )

;; Ray-casting algorithm (http://en.wikipedia.org/wiki/Point_in_polygon)
;; (http://alienryderflex.com/polygon/)
;
(define (intersects?-pip pi pj lat lon)
  (let ((lati (geopoint-latitude pi))
        (loni (geopoint-longitude pi))
        (latj (geopoint-latitude pj))
        (lonj (geopoint-longitude pj)) )
    (and (or (and (< lati lat) (<= lat latj ) ) (and (< latj lat) (<= lat lati)))
         (or (<= loni lon) (<= lonj lon))
         (< (+ loni (* (/ (- lat lati) (- latj lati)) (- lonj loni))) lon)) ) )

) ;module geopoint-utils
